from typing import Dict, Tuple


def _ceil100(x):
    if x % 100 > 0:
        x += 100 - x % 100
    return x


no_ron = {
    (1, 20), (1, 25),
    (2, 20),
    (3, 20),
    (4, 20)
}

no_tsumo = {
    (1, 20), (1, 25), (1, 110),
    (2, 25),
}


def _calc_parent_point(han: int, hu: int):
    a = hu * (2 ** (han + 2))
    if a > 2000:
        a = 2000

    ron = _ceil100(6 * a)
    tsumo = _ceil100(2 * a)

    if (han, hu) in no_ron:
        ron = 0
    if (han, hu) in no_tsumo:
        tsumo = 0

    return ron, tsumo


def build_han_hu_to_parent_point(mapping):
    """
    (番, 符) -> (荣和点数, 自摸各家点数)，5番及以上的结果存到20符里
    :param mapping:
    :return:
    """
    for han in range(1, 5):
        for hu in range(20, 120, 10):
            if not ((han, hu) in no_ron and (han, hu) in no_tsumo):
                mapping[han, hu] = _calc_parent_point(han, hu)

    for han in range(2, 4):
        mapping[han, 25] = _calc_parent_point(han, 25)

    mapping[5, 20] = 12000, 4000
    mapping[6, 20] = 18000, 6000
    mapping[7, 20] = 18000, 6000
    mapping[8, 20] = 24000, 8000
    mapping[9, 20] = 24000, 8000
    mapping[10, 20] = 24000, 8000
    mapping[11, 20] = 36000, 12000
    mapping[12, 20] = 36000, 12000
    mapping[13, 20] = 48000, 16000


def _calc_child_point(han: int, hu: int):
    a = hu * (2 ** (han + 2))
    if a > 2000:
        a = 2000

    ron = _ceil100(4 * a)
    tsumo_parent = _ceil100(2 * a)
    tsumo_child = _ceil100(a)

    if (han, hu) in no_ron:
        ron = 0
    if (han, hu) in no_tsumo:
        tsumo_parent = 0
        tsumo_child = 0

    return ron, tsumo_parent, tsumo_child


def build_han_hu_to_child_point(mapping):
    """
    (番, 符) -> (荣和点数, 自摸庄家点数, 自摸闲家点数)，5番及以上的结果存到20符里
    :param mapping:
    :return:
    """
    for han in range(1, 5):
        for hu in range(20, 120, 10):
            if not ((han, hu) in no_ron and (han, hu) in no_tsumo):
                mapping[han, hu] = _calc_child_point(han, hu)

    for han in range(2, 4):
        mapping[han, 25] = _calc_child_point(han, 25)

    mapping[5, 20] = 8000, 4000, 2000
    mapping[6, 20] = 12000, 6000, 3000
    mapping[7, 20] = 12000, 6000, 3000
    mapping[8, 20] = 16000, 8000, 4000
    mapping[9, 20] = 16000, 8000, 4000
    mapping[10, 20] = 16000, 8000, 4000
    mapping[11, 20] = 24000, 12000, 6000
    mapping[12, 20] = 24000, 12000, 6000
    mapping[13, 20] = 32000, 16000, 8000


parent_han_hu_mapping: Dict[Tuple[int, int], Tuple[int, int, int]] = {}
child_han_hu_mapping: Dict[Tuple[int, int], Tuple[int, int]] = {}

build_han_hu_to_parent_point(parent_han_hu_mapping)
build_han_hu_to_child_point(child_han_hu_mapping)


def get_parent_point_by_han_hu(han: int, hu: int):
    """
    获取亲家X番Y符的点数

    :param han: 番
    :param hu: 符
    :return: (荣和点数, 自摸各家点数)
    """
    if han >= 13:
        han = 13
    if han >= 5:
        hu = 20
    if (han, hu) not in parent_han_hu_mapping:
        raise ValueError(f"invalid arguments: han={han}, hu={hu}")

    return parent_han_hu_mapping[han, hu]


def get_child_point_by_han_hu(han: int, hu: int):
    """
    获取子家X番Y符的点数

    :param han: 番
    :param hu: 符
    :return: (荣和点数, 自摸庄家点数, 自摸闲家点数)
    """
    if han >= 13:
        han = 13
    if han >= 5:
        hu = 20
    if (han, hu) not in child_han_hu_mapping:
        raise ValueError(f"invalid arguments: han={han}, hu={hu}")

    return child_han_hu_mapping[han, hu]


__all__ = ("get_parent_point_by_han_hu", "get_child_point_by_han_hu",)



from typing import Sequence

from pydantic.fields import Field
from pydantic.main import BaseModel

from mahjong_utils.models.furo import Furo, Kan
from mahjong_utils.models.hand_pattern import HandPattern
from mahjong_utils.models.tile import Tile


class Hand(BaseModel):
    tiles: Sequence[Tile]
    furo: Sequence[Furo] = Field(default_factory=tuple)
    patterns: Sequence[HandPattern]

    @property
    def menzen(self) -> bool:
        for fr in self.furo:
            if not isinstance(fr, Kan) or not fr.ankan:
                return False
        return True



from typing import Optional, Set, List
from typing import Tuple

from pydantic import BaseModel, Field

from mahjong_utils.models.furo import Furo
from mahjong_utils.models.hora_hand_pattern import HoraHandPattern, build_hora_hand
from mahjong_utils.models.tile import Tile
from mahjong_utils.models.wind import Wind
from mahjong_utils.point_by_han_hu import get_parent_point_by_han_hu, get_child_point_by_han_hu
from mahjong_utils.shanten import ShantenResult, shanten, UnionShantenResult
from mahjong_utils.yaku import Yaku
from mahjong_utils.yaku.check import check_yaku


def _calc_han(yaku: Set[Yaku], menzen: bool) -> int:
    if menzen:
        han = sum(map(lambda y: y.han, yaku))
    else:
        han = sum(map(lambda y: y.han - y.furo_loss, yaku))
    return han


class Hora(BaseModel):
    pattern: HoraHandPattern
    dora: int = 0
    extra_yaku: Optional[Set[Yaku]]
    han: int = 0  # 在__init__中计算
    yaku: Set[Yaku] = Field(default_factory=set)  # 在__init__中计算

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.yaku = check_yaku(self.pattern, self.extra_yaku)
        self.han = _calc_han(self.yaku, self.pattern.menzen)
        if self.han > 0:
            self.han += self.dora

    @property
    def hu(self) -> int:
        return self.pattern.hu

    @property
    def tsumo(self) -> bool:
        return self.pattern.tsumo

    @property
    def self_wind(self) -> Optional[Wind]:
        return self.pattern.self_wind

    @property
    def round_wind(self) -> Optional[Wind]:
        return self.pattern.round_wind

    @property
    def agari(self) -> Tile:
        return self.pattern.agari

    @property
    def parent_point(self) -> Tuple[int, int]:
        """
        亲家和牌点数

        :return: (荣和点数, 自摸各家点数)
        """
        if len(self.yaku) == 0:
            return 0, 0
        elif next(iter(self.yaku)).is_yakuman:
            times = 0
            for yaku in self.yaku:
                times += yaku.han // 13

            ans = get_parent_point_by_han_hu(self.han, self.pattern.hu)
            return ans[0] * times, ans[1] * times
        else:
            return get_parent_point_by_han_hu(self.han, self.pattern.hu)

    @property
    def child_point(self) -> Tuple[int, int, int]:
        """
        子家X番Y符的点数

        :return: (荣和点数, 自摸庄家点数, 自摸闲家点数)
        """
        if len(self.yaku) == 0:
            return 0, 0, 0
        elif next(iter(self.yaku)).is_yakuman:
            times = 0
            for yaku in self.yaku:
                times += yaku.han // 13

            ans = get_child_point_by_han_hu(13, 20)
            return ans[0] * times, ans[1] * times, ans[2] * times
        else:
            return get_child_point_by_han_hu(self.han, self.pattern.hu)


def build_hora(tiles: List[Tile], furo: Optional[List[Furo]], agari: Tile,
               tsumo: bool,
               *,
               dora: int = 0,
               self_wind: Optional[Wind] = None, round_wind: Optional[Wind] = None,
               extra_yaku: Optional[Set[Yaku]] = None) -> Hora:
    """
    根据手牌构造Hora

    :param tiles: 手牌
    :param furo: 副露
    :param agari: 和牌
    :param tsumo: 是否自摸
    :param dora: 宝牌数
    :param self_wind: 自风
    :param round_wind: 场风
    :param extra_yaku: 额外役
    :return: Hora
    """
    if furo is None:
        furo = []

    k = len(tiles) // 3 + len(furo)
    if k != 4:
        raise ValueError("invalid length of _tiles")

    shanten_result = shanten(tiles, furo)
    return build_hora_from_shanten_result(shanten_result, agari, tsumo,
                                          dora=dora, self_wind=self_wind, round_wind=round_wind,
                                          extra_yaku=extra_yaku)


def build_hora_from_shanten_result(
        shanten_result: ShantenResult,
        agari: Tile,
        tsumo: bool,
        *, dora: int = 0,
        self_wind: Optional[Wind] = None, round_wind: Optional[Wind] = None,
        extra_yaku: Optional[Set[Yaku]] = None
) -> Hora:
    """
    根据向听分析结果构造Hora

    :param shanten_result: 向听分析结果
    :param agari: 和牌
    :param tsumo: 是否自摸
    :param dora: 宝牌数
    :param self_wind: 自风
    :param round_wind: 场风
    :param extra_yaku: 额外役
    :return: Hora
    """
    if not shanten_result.with_got:
        raise ValueError("shanten_result without got")
    if shanten_result.shanten != -1:
        raise ValueError("shanten != -1")

    possible_hora = []

    patterns = []

    if isinstance(shanten_result, UnionShantenResult):
        if shanten_result.regular.shanten == -1:
            patterns += shanten_result.regular.hand.patterns
        if shanten_result.chitoi is not None and shanten_result.chitoi.shanten == -1:
            patterns += shanten_result.chitoi.hand.patterns
        if shanten_result.kokushi is not None and shanten_result.kokushi.shanten == -1:
            patterns += shanten_result.kokushi.hand.patterns
    else:
        patterns = shanten_result.hand.patterns

    for pat in patterns:
        for hora_hand in build_hora_hand(pat, agari, tsumo, self_wind, round_wind):
            hora = Hora(pattern=hora_hand, dora=dora, extra_yaku=extra_yaku)
            possible_hora.append(hora)

    best_hora = possible_hora[0]
    for hora in possible_hora:
        if hora.han > best_hora.han or (hora.han == best_hora.han and hora.hu > best_hora.hu):
            best_hora = hora
    return best_hora


__all__ = ("Hora", "build_hora", "build_hora_from_shanten_result")



from typing import Set, Optional

from mahjong_utils.models.hora_hand_pattern import HoraHandPattern
from mahjong_utils.yaku import Yaku
from mahjong_utils.yaku.common import all_common_yaku
from mahjong_utils.yaku.yakuman import all_yakuman


def check_yaku(hand: HoraHandPattern, extra_yaku: Optional[Set[Yaku]] = None) -> Set[Yaku]:
    ans = set()
    has_yakuman = False

    for yakuman in all_yakuman:
        # 副露无效的役直接跳过
        if (hand.menzen or yakuman.furo_loss != yakuman.han) and yakuman(hand):
            has_yakuman = True
            ans.add(yakuman)

    if extra_yaku is not None:
        for yakuman in filter(lambda x: x in all_yakuman, extra_yaku):
            has_yakuman = True
            ans.add(yakuman)

    if not has_yakuman:
        for yaku in all_common_yaku:
            if (hand.menzen or yaku.furo_loss != yaku.han) and yaku(hand):
                ans.add(yaku)
        if extra_yaku is not None:
            for yaku in filter(lambda x: x not in all_yakuman, extra_yaku):
                ans.add(yaku)

    return ans



from abc import ABC
from typing import List
from typing import Optional, Iterable

from pydantic import Field, root_validator

from mahjong_utils.internal.attr_dict import attr_dict
from mahjong_utils.internal.hand_utils import regular_pattern_after_discard
from mahjong_utils.models.furo import Pon, Kan
from mahjong_utils.models.hand_pattern import RegularHandPattern, HandPattern, ChitoiHandPattern, KokushiHandPattern
from mahjong_utils.models.mentsu import Kotsu
from mahjong_utils.models.tatsu import Penchan, Kanchan, Tatsu
from mahjong_utils.models.tile import Tile
from mahjong_utils.models.tile import all_yaochu
from mahjong_utils.models.tile import is_yaochu
from mahjong_utils.models.tile_type import TileType
from mahjong_utils.models.wind import Wind


class HoraHandPattern(HandPattern, ABC):
    agari: Tile
    tsumo: bool
    hu: int
    self_wind: Optional[Wind]
    round_wind: Optional[Wind]


class RegularHoraHandPattern(HoraHandPattern, RegularHandPattern):
    agari_tatsu: Optional[Tatsu]
    hu: int = 0  # 在__init__中计算

    k: int = 4
    jyantou: Tile

    @root_validator
    def validate_k(cls, values):
        assert len(values["menzen_mentsu"]) + len(values["furo"]) == 4
        return values

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.hu = self._calc_hu()

    def _calc_hu(self) -> int:
        ans = 20

        # 单骑、边张、坎张听牌
        if self.agari_tatsu is None or isinstance(self.agari_tatsu, Penchan) or isinstance(self.agari_tatsu, Kanchan):
            ans += 2

        # 明刻、杠
        for fr in self.furo:
            if isinstance(fr, Pon):
                if is_yaochu(fr.tile):
                    ans += 4
                else:
                    ans += 2
            elif isinstance(fr, Kan):
                if not fr.ankan:
                    if is_yaochu(fr.tile):
                        ans += 16
                    else:
                        ans += 8
                else:
                    if is_yaochu(fr.tile):
                        ans += 32
                    else:
                        ans += 16

        # 暗刻（不含暗杠）
        for mt in self.menzen_mentsu:
            if isinstance(mt, Kotsu):
                if is_yaochu(mt.tile):
                    ans += 8
                else:
                    ans += 4

        # 对碰荣和时该刻子计为明刻
        if isinstance(self.agari_tatsu, Kotsu):
            ans -= 2

        # 役牌雀头（连风算4符）
        if self.self_wind is not None and self.jyantou == self.self_wind.tile:
            ans += 2
        if self.round_wind is not None and self.jyantou == self.round_wind.tile:
            ans += 2
        if self.jyantou.tile_type == TileType.Z and self.jyantou.num >= 5:
            ans += 2

        # 门清荣和
        if self.menzen and not self.tsumo:
            ans += 10

        # 非平和自摸
        if ans != 20 and self.tsumo:
            ans += 2

        # 非门清最低30符
        if not self.menzen and ans < 30:
            ans = 30

        # 切上
        if ans % 10 != 0:
            ans += 10 - ans % 10

        return ans


class ChitoiHoraHandPattern(HoraHandPattern, ChitoiHandPattern):
    hu: int = 25


class KokushiHoraHandPattern(HoraHandPattern, KokushiHandPattern):
    hu: int = 20

    yaochu: List[Tile] = Field(default_factory=lambda: list(all_yaochu))
    repeated: Tile
    remaining: List[Tile] = Field(default_factory=list)

    @property
    def thirteen_waiting(self) -> bool:
        return self.agari == self.repeated


def _build_regular_hora_hand_pattern(
        pattern: RegularHandPattern,
        agari: Tile,
        tsumo: bool,
        self_wind: Optional[Wind] = None,
        round_wind: Optional[Wind] = None,
        *, with_got: bool = True
) -> Iterable[RegularHoraHandPattern]:
    if not with_got:
        if pattern.jyantou is not None:
            yield RegularHoraHandPattern(agari=agari,
                                         tsumo=tsumo,
                                         self_wind=self_wind,
                                         round_wind=round_wind,
                                         agari_tatsu=pattern.tatsu[0],
                                         jyantou=pattern.jyantou,
                                         menzen_mentsu=pattern.menzen_mentsu + (pattern.tatsu[0].with_waiting(agari),),
                                         furo=pattern.furo)
        else:
            yield RegularHoraHandPattern(agari=agari,
                                         tsumo=tsumo,
                                         self_wind=self_wind,
                                         round_wind=round_wind,
                                         agari_tatsu=None,
                                         jyantou=agari,
                                         menzen_mentsu=pattern.menzen_mentsu,
                                         furo=pattern.furo)
    else:
        for pat_after_discard in regular_pattern_after_discard(pattern, agari):
            for h in _build_regular_hora_hand_pattern(pat_after_discard,
                                                      agari, tsumo,
                                                      self_wind, round_wind,
                                                      with_got=False):
                yield h


def _build_chitoi_hora_hand(
        pattern: ChitoiHandPattern,
        agari: Tile,
        tsumo: bool,
        self_wind: Optional[Wind] = None,
        round_wind: Optional[Wind] = None
) -> Iterable[ChitoiHoraHandPattern]:
    yield ChitoiHoraHandPattern(agari=agari,
                                tsumo=tsumo,
                                self_wind=self_wind,
                                round_wind=round_wind,
                                pairs=pattern.pairs)


def _build_kokushi_hora_hand(
        pattern: KokushiHandPattern,
        agari: Tile,
        tsumo: bool,
        self_wind: Optional[Wind] = None,
        round_wind: Optional[Wind] = None
) -> Iterable[KokushiHoraHandPattern]:
    yield KokushiHoraHandPattern(agari=agari,
                                 tsumo=tsumo,
                                 self_wind=self_wind,
                                 round_wind=round_wind,
                                 repeated=pattern.repeated)


def build_hora_hand(
        pattern: HandPattern,
        agari: Tile,
        tsumo: bool,
        self_wind: Optional[Wind] = None,
        round_wind: Optional[Wind] = None
) -> Iterable[HoraHandPattern]:
    """
    根据给定Hand构造HoraHand。返回所有可能的HoraHand。

    :param pattern: HandPattern
    :param agari: 和牌
    :param tsumo: 是否自摸
    :param self_wind: 自风
    :param round_wind: 场风
    :return: 当Hand为摸牌状态时，返回所有可能的HoraHand。否则返回一个对应的HoraHand。
    """
    if isinstance(pattern, RegularHandPattern):
        return _build_regular_hora_hand_pattern(pattern, agari, tsumo, self_wind, round_wind)
    elif isinstance(pattern, ChitoiHandPattern):
        return _build_chitoi_hora_hand(pattern, agari, tsumo, self_wind, round_wind)
    elif isinstance(pattern, KokushiHandPattern):
        return _build_kokushi_hora_hand(pattern, agari, tsumo, self_wind, round_wind)
    else:
        raise TypeError(f"unexpected type of pattern: {type(pattern)}")



from typing import List, Callable, Tuple, Generator, Sequence

from mahjong_utils.internal.tile_type_mapping import tile_type_index_mapping, tile_type_reversed_index_mapping
from mahjong_utils.internal.utils.bit import generate_k_bit_number
from mahjong_utils.models.furo import Furo
from mahjong_utils.models.hand_pattern import RegularHandPattern
from mahjong_utils.models.mentsu import Kotsu, Shuntsu, Mentsu
from mahjong_utils.models.tatsu import Tatsu, Toitsu, Kanchan, Ryanmen, Penchan
from mahjong_utils.models.tile import Tile, tile
from mahjong_utils.models.tile_type import TileType


def _encode(t: Tile) -> int:
    return tile_type_index_mapping[t.tile_type] * 9 + t.real_num - 1


def _decode(code: int) -> Tile:
    tile_type = tile_type_reversed_index_mapping[code // 9]
    num = code % 9 + 1
    return tile(tile_type, num)


class RegularHandPatternSearcher:
    def __init__(self, tiles: Sequence[Tile], furo: Sequence[Furo],
                 callback: Callable[[RegularHandPattern], None]):
        self._callback = callback

        if not isinstance(furo, tuple):
            furo = tuple(furo)
        self._furo = furo

        self._count = [0] * (3 * 9 + 7)
        for t in tiles:
            self._count[_encode(t)] += 1

        self._n = len(tiles)
        self._k = self._n // 3

        self._mentsu: List[Mentsu] = []
        self._tatsu: List[Tatsu] = []

    def run(self):
        self._dfs_kotsu()

    def _dfs_kotsu(self, begin=0):
        # begin用于限制从哪张牌开始枚举（下同）
        # 其目的是避免搜索时按不同顺序取了相同的刻字，优化性能

        if self._n >= 3:
            for i in range(begin, 3 * 9 + 7):
                if self._count[i] >= 3:
                    self._n -= 3
                    self._count[i] -= 3
                    self._mentsu.append(Kotsu(_decode(i)))
                    self._dfs_kotsu(i)
                    self._n += 3
                    self._count[i] += 3
                    self._mentsu.pop()

        self._dfs_shuntsu()

    def _dfs_shuntsu(self, begin=0):
        if self._n >= 3:
            for i in range(begin // 9, 3):  # m/p/s
                for j in range(7):  # 1~7
                    x = i * 9 + j
                    y = x + 1
                    z = x + 2

                    if x < begin:
                        continue

                    if self._count[x] > 0 and self._count[y] > 0 and self._count[z] > 0:
                        self._n -= 3
                        self._count[x] -= 1
                        self._count[y] -= 1
                        self._count[z] -= 1
                        self._mentsu.append(Shuntsu(_decode(x)))
                        self._dfs_shuntsu(x)
                        self._n += 3
                        self._count[x] += 1
                        self._count[y] += 1
                        self._count[z] += 1
                        self._mentsu.pop()

        self._dfs_tatsu()

    def _dfs_tatsu(self, begin=0, tatsu_type_limitation=0):
        # tatsu_type_limitation用于限制能够取什么样的以begin为第一张牌的搭子（0可以取所有类型，1不可以取对子、2不可以取对子和坎张、3不可以取对子坎张两面）
        # 其目的是避免搜索时按不同顺序取了相同的搭子，优化性能
        # 故当for循环执行了一趟以后就将tatsu_type_limitation置0

        taken = False

        if self._n >= 2:
            for i in range(begin, 3 * 9 + 7):
                t = _decode(i)

                # toitsu
                if tatsu_type_limitation == 0 and self._count[i] >= 2:
                    taken = True
                    self._n -= 2
                    self._count[i] -= 2
                    self._tatsu.append(Toitsu(_decode(i)))
                    self._dfs_tatsu(i, 0)
                    self._n += 2
                    self._count[i] += 2
                    self._tatsu.pop()

                # kanchan
                if tatsu_type_limitation <= 1 and t.tile_type != TileType.Z and t.num <= 7:
                    j = i + 2
                    if self._count[i] > 0 and self._count[j] > 0:
                        taken = True
                        self._n -= 2
                        self._count[i] -= 1
                        self._count[j] -= 1
                        self._tatsu.append(Kanchan(t))
                        self._dfs_tatsu(i, 1)
                        self._n += 2
                        self._count[i] += 1
                        self._count[j] += 1
                        self._tatsu.pop()

                # ryanmen
                if tatsu_type_limitation <= 2 and 2 <= t.num <= 7 and t.tile_type != TileType.Z:
                    j = i + 1
                    if self._count[i] > 0 and self._count[j] > 0:
                        taken = True
                        self._n -= 2
                        self._count[i] -= 1
                        self._count[j] -= 1
                        self._tatsu.append(Ryanmen(t))
                        self._dfs_tatsu(i, 2)
                        self._n += 2
                        self._count[i] += 1
                        self._count[j] += 1
                        self._tatsu.pop()

                # penchan
                if tatsu_type_limitation <= 3 and t.tile_type != TileType.Z and (t.num == 1 or t.num == 8):
                    j = i + 1
                    if self._count[i] > 0 and self._count[j] > 0:
                        taken = True
                        self._n -= 2
                        self._count[i] -= 1
                        self._count[j] -= 1
                        self._tatsu.append(Penchan(t))
                        self._dfs_tatsu(i, 3)
                        self._n += 2
                        self._count[i] += 1
                        self._count[j] += 1
                        self._tatsu.pop()

                tatsu_type_limitation = 0

        if not taken:
            self._on_result()

    def _on_result(self):
        for hand in self._normalize():
            self._callback(hand)

    def _normalize(self):
        # 将搜索结果处理为（雀头，面子，搭子，浮牌）的形式，且面子数+搭子数不超过k
        remaining = []
        for i in range(3 * 9 + 7):
            if self._count[i] > 0:
                t = _decode(i)
                for j in range(self._count[i]):
                    remaining.append(t)
        remaining = tuple(remaining)

        has_toitsu = False

        # enumerate jyantou
        for i, tt in enumerate(self._tatsu):
            if isinstance(tt, Toitsu):
                has_toitsu = True

                remaining_tatsu = self._tatsu[0:i] + self._tatsu[i + 1:]

                for tatsu_chosen, tatsu_not_chosen_as_tiles in self._choose_tatsu(self._k - len(self._mentsu),
                                                                                  remaining_tatsu):
                    yield RegularHandPattern(k=self._k + len(self._furo),
                                             jyantou=tt.first,
                                             menzen_mentsu=tuple(self._mentsu),
                                             furo=self._furo,
                                             tatsu=tatsu_chosen,
                                             remaining=remaining + tatsu_not_chosen_as_tiles)

        if not has_toitsu:
            for tatsu_chosen, tatsu_not_chosen_as_tiles in self._choose_tatsu(self._k - len(self._mentsu), self._tatsu):
                yield RegularHandPattern(k=self._k + len(self._furo),
                                         jyantou=None,
                                         menzen_mentsu=tuple(self._mentsu),
                                         furo=self._furo,
                                         tatsu=tatsu_chosen,
                                         remaining=remaining + tatsu_not_chosen_as_tiles)

    @staticmethod
    def _choose_tatsu(k: int, tatsu: List[Tatsu]) -> Generator[Tuple[Tuple[Tatsu], Tuple[Tile]], None, None]:
        # 选择k个搭子
        if k >= len(tatsu):
            yield tuple(tatsu), tuple()
        elif k == 0:
            tiles = []
            for tt in tatsu:
                tiles.append(tt.first)
                tiles.append(tt.second)
            yield tuple(), tuple(tiles)
        else:
            maximum = 1 << len(tatsu)
            for mask in generate_k_bit_number(k):
                if mask >= maximum:
                    break

                tatsu_chosen = []
                tatsu_not_chosen_as_tiles = []

                for i in range(len(tatsu)):
                    if mask & (1 << i):
                        tatsu_chosen.append(tatsu[i])
                    else:
                        tatsu_not_chosen_as_tiles.append(tatsu[i].first)
                        tatsu_not_chosen_as_tiles.append(tatsu[i].second)

                yield tuple(tatsu_chosen), tuple(tatsu_not_chosen_as_tiles)


def regular_hand_pattern_search(tiles: Sequence[Tile], furo: Sequence[Furo]) -> List[RegularHandPattern]:
    patterns = []

    def callback(pattern: RegularHandPattern):
        patterns.append(pattern)

    searcher = RegularHandPatternSearcher(tiles, furo, callback)
    searcher.run()

    return patterns


__all__ = ("regular_hand_pattern_search",)



from typing import Iterable, Set

from mahjong_utils.internal.tile_cling import tile_cling
from mahjong_utils.models.hand_pattern import RegularHandPattern, ChitoiHandPattern, KokushiHandPattern
from mahjong_utils.models.tatsu import parse_tatsu
from mahjong_utils.models.tile import Tile, tile




def calc_regular_advance(pattern: RegularHandPattern) -> Set[Tile]:
    """
    计算进张

    :param pattern:
    :return:
    """
    advance = set()

    # 搭子的进张
    for tt in pattern.tatsu:
        advance |= tt.waiting

    # 浮张的靠张
    if len(pattern.furo) + len(pattern.menzen_mentsu) + len(pattern.tatsu) < pattern.k:
        for t in pattern.remaining:
            advance |= tile_cling[t]

    # 无雀头
    if pattern.jyantou is None:
        for t in pattern.remaining:
            advance.add(t)

    assert tile("0m") not in advance
    assert tile("0p") not in advance
    assert tile("0s") not in advance

    return advance


def regular_pattern_after_discard(pattern: RegularHandPattern, discard: Tile) -> Iterable[RegularHandPattern]:
    """

    :param pattern:
    :param discard:
    :return:
    """
    # 扣掉雀头
    if pattern.jyantou == discard:
        new_hand = pattern.copy(update={
            "jyantou": None,
            "remaining": pattern.remaining + (discard,)
        })
        yield new_hand

    # 扣掉面子
    for i, mt in enumerate(pattern.menzen_mentsu):
        try:
            tatsu = mt.after_discard(discard)
        except ValueError:
            continue

        new_hand = pattern.copy(update={
            "menzen_mentsu": pattern.menzen_mentsu[:i] + pattern.menzen_mentsu[i + 1:],
            "tatsu": pattern.tatsu + (tatsu,)
        })
        yield new_hand

    # 扣掉搭子
    for i, tt in enumerate(pattern.tatsu):
        if discard == tt.first:
            new_hand = pattern.copy(update={
                "tatsu": pattern.tatsu[:i] + pattern.tatsu[i + 1:],
                "remaining": pattern.remaining + (tt.second,)
            })
            yield new_hand
        elif discard == tt.second:
            new_hand = pattern.copy(update={
                "tatsu": pattern.tatsu[:i] + pattern.tatsu[i + 1:],
                "remaining": pattern.remaining + (tt.first,)
            })
            yield new_hand

    # 扣掉浮张
    try:
        idx = pattern.remaining.index(discard)
    except ValueError:
        idx = -1

    if idx != -1:
        new_hand = pattern.copy(update={
            "remaining": pattern.remaining[:idx] + pattern.remaining[idx + 1:]
        })
        yield new_hand


def regular_pattern_after_advance(pattern: RegularHandPattern, advance: Tile) -> Iterable[RegularHandPattern]:
    # 搭子的进张
    for i, tt in enumerate(pattern.tatsu):
        if advance in tt.waiting:
            new_hand = pattern.copy(update={
                "menzen_mentsu": pattern.menzen_mentsu + (tt.with_waiting(advance),),
                "tatsu": pattern.tatsu[:i] + pattern.tatsu[i + 1:]
            })
            yield new_hand

    # 浮张的靠张
    if len(pattern.furo) + len(pattern.menzen_mentsu) + len(pattern.tatsu) < pattern.k:
        for i, t in enumerate(pattern.remaining):
            if advance in tile_cling[t]:
                new_hand = pattern.copy(update={
                    "tatsu": pattern.tatsu + (parse_tatsu((advance, t)),),
                    "remaining": pattern.remaining[:i] + pattern.remaining[i + 1:]
                })
                yield new_hand
                break

    # 无雀头
    if pattern.jyantou is None:
        for i, t in enumerate(pattern.remaining):
            if t == advance:
                new_hand = pattern.copy(update={
                    "jyantou": t,
                    "remaining": pattern.remaining[:i] + pattern.remaining[i + 1:]
                })
                yield new_hand
                break



from mahjong_utils.models.tile_type import TileType

tile_type_index_mapping = {TileType.M: 0, TileType.P: 1, TileType.S: 2, TileType.Z: 3}
tile_type_reversed_index_mapping = [TileType.M, TileType.P, TileType.S, TileType.Z]
__all__ = ("tile_type_index_mapping", "tile_type_reversed_index_mapping")



from typing import Dict, Set

from mahjong_utils.internal.tile_type_mapping import tile_type_reversed_index_mapping
from mahjong_utils.models.tile import Tile, tile
from mahjong_utils.models.tile_type import TileType

tile_cling: Dict[Tile, Set[Tile]] = {}

for i in range(3):
    tile_type = tile_type_reversed_index_mapping[i]

    for j in range(1, 10):
        t = tile(tile_type, j)
        tile_cling[t] = set()

        for k in {-2, -1, 0, 1, 2}:
            if 1 <= j + k <= 9:
                tile_cling[t].add(tile(tile_type, j + k))

    tile_cling[tile(tile_type, 0)] = tile_cling[tile(tile_type, 5)]

for j in range(1, 8):
    t = tile(TileType.Z, j)
    tile_cling[t] = {t}

__all__ = ("tile_cling",)



from typing import Set

from mahjong_utils.models.hora_hand_pattern import HoraHandPattern, KokushiHoraHandPattern
from mahjong_utils.models.tile import is_z, parse_tiles
from mahjong_utils.yaku import _yaku, Yaku
from mahjong_utils.yaku.checker_factory import kantsu_series_checker_factory, anko_series_checker_factory, \
    yaochu_series_checker_factory, \
    sangen_series_checker_factory, sushi_series_checker_factory, churen_series_checker_factory
from mahjong_utils.yaku.extra import tenhou, chihou


@_yaku(13, 13, True)
def kokushi(hora_hand: HoraHandPattern) -> bool:
    return isinstance(hora_hand, KokushiHoraHandPattern) and not hora_hand.thirteen_waiting


suanko = Yaku("suanko", 13, 13, True, anko_series_checker_factory(4, False))

daisangen = Yaku("daisangen", 13, 0, True, sangen_series_checker_factory(3, False))


@_yaku(13, 0, True)
def tsuiso(hora_hand: HoraHandPattern) -> bool:
    for t in hora_hand.tiles:
        if not is_z(t):
            return False
    return True


shousushi = Yaku("shousushi", 13, 0, True, sushi_series_checker_factory(3, True))

lyu = {*parse_tiles("23468s6z")}


@_yaku(13, 0, True)
def lyuiso(hora_hand: HoraHandPattern) -> bool:
    for t in hora_hand.tiles:
        if t not in lyu:
            return False
    return True


chinroto = Yaku("chinroto", 13, 0, True, yaochu_series_checker_factory(False, False))

sukantsu = Yaku("sukantsu", 13, 0, True, kantsu_series_checker_factory(4))

churen = Yaku("churen", 13, 13, True, churen_series_checker_factory(False))

daisushi = Yaku("daisushi", 26, 0, True, sushi_series_checker_factory(4, False))

churen9 = Yaku("churen9", 26, 26, True, churen_series_checker_factory(True))

suanko_tanki = Yaku("suanko_tanki", 26, 26, True, anko_series_checker_factory(4, True))


@_yaku(26, 26, True)
def kokushi13(hora_hand: HoraHandPattern) -> bool:
    return isinstance(hora_hand, KokushiHoraHandPattern) and hora_hand.thirteen_waiting


all_yakuman: Set[Yaku] = {
    tenhou, chihou, kokushi, suanko, daisangen, tsuiso, shousushi, lyuiso, chinroto, sukantsu, churen,
    daisushi, churen9, suanko_tanki, kokushi13
}

__all__ = (
    "all_yakuman",
    "tenhou", "chihou", "kokushi", "suanko", "daisangen", "tsuiso", "shousushi", "lyuiso", "chinroto", "sukantsu", "churen",
    "daisushi", "churen9", "suanko_tanki", "kokushi13"
)



from typing import Optional

from mahjong_utils.internal.tile_type_mapping import tile_type_index_mapping
from mahjong_utils.models.furo import Kan
from mahjong_utils.models.hora_hand_pattern import HoraHandPattern, RegularHoraHandPattern
from mahjong_utils.models.mentsu import Shuntsu, Kotsu
from mahjong_utils.models.tatsu import Toitsu
from mahjong_utils.models.tile import is_yaochu, Tile, is_sangen, is_z, is_wind


def yakuhai_checker_factory(tile: Tile):
    def checker(hora_hand: HoraHandPattern) -> bool:
        if not isinstance(hora_hand, RegularHoraHandPattern):
            return False

        for kt in hora_hand.kotsu:
            if kt.tile == tile:
                return True
        return False

    return checker


def yaochu_series_checker_factory(shuntsu: bool, z: bool):
    """
    幺九系列checker工厂（混全、混老头、纯全、清老头）
    :param shuntsu: True 必须含顺子，False 必须不含顺子
    :param z: True 必须含字牌，False 必须不含字牌
    :return: checker
    """

    def checker(hora_hand: HoraHandPattern) -> bool:
        if not isinstance(hora_hand, RegularHoraHandPattern):
            return False

        if not is_yaochu(hora_hand.jyantou):
            return False

        shuntsu_cnt = 0
        kotsu_cnt = 0
        has_z = is_z(hora_hand.jyantou)

        for st in hora_hand.shuntsu:
            if st.tile.num != 1 and st.tile.num != 7:
                return False
            shuntsu_cnt += 1

        for kt in hora_hand.kotsu:
            if not is_yaochu(kt.tile):
                return False
            kotsu_cnt += 1
            if is_z(kt.tile):
                has_z = True

        return (shuntsu and shuntsu_cnt != 0 or not shuntsu and shuntsu_cnt == 0) and \
               (z and has_z or not z and not has_z)

    return checker


def itsu_series_checker_factory(z: bool):
    """
    一色系列checker工厂（混一色、清一色）
    :param z: True 必须含字牌，False 必须不含字牌
    :return: checker
    """

    def checker(hora_hand: HoraHandPattern) -> bool:
        cnt = [0] * 4

        for t in hora_hand.tiles:
            cnt[tile_type_index_mapping[t.tile_type]] += 1

        zero_cnt = 0
        for i in range(3):
            if cnt[i] == 0:
                zero_cnt += 1

        return zero_cnt == 2 and (
                z and cnt[3] != 0 or
                not z and cnt[3] == 0
        )

    return checker


def peko_series_checker_factory(peko_count: int):
    """
    杯口系列checker工厂（一杯口、两杯口）
    :param peko_count: x杯口
    :return: checker
    """

    def checker(hora_hand: HoraHandPattern) -> bool:
        if not isinstance(hora_hand, RegularHoraHandPattern) or not hora_hand.menzen:
            return False

        cnt = 0
        for i in range(len(hora_hand.menzen_mentsu)):
            for j in range(i + 1, len(hora_hand.menzen_mentsu)):
                if isinstance(hora_hand.menzen_mentsu[i], Shuntsu) \
                        and hora_hand.menzen_mentsu[i] == hora_hand.menzen_mentsu[j]:
                    cnt += 1
        return cnt == peko_count

    return checker


def anko_series_checker_factory(anko_count: int, tanki: Optional[bool] = None):
    """
    暗刻系列checker工厂（三暗刻、四暗刻、四暗刻单骑）
    :param anko_count: x暗刻
    :param tanki: True 必须单骑，False 必须非单骑，None 无所谓
    :return: checker
    """

    def checker(hora_hand: HoraHandPattern) -> bool:
        if not isinstance(hora_hand, RegularHoraHandPattern):
            return False

        anko = 0
        for mt in hora_hand.anko:
            if isinstance(mt, Kotsu):
                anko += 1

        if isinstance(hora_hand.agari_tatsu, Toitsu) and not hora_hand.tsumo:  # 双碰听牌荣和，算一个明刻
            anko -= 1

        if tanki is None:
            return anko == anko_count
        elif tanki:
            return anko == anko_count and hora_hand.agari_tatsu is None
        else:
            return anko == anko_count and hora_hand.agari_tatsu is not None

    return checker


def kantsu_series_checker_factory(kan_count: int):
    """
    杠子系列checker工厂（三杠子、四杠子）
    :param kan_count: x杠子
    :return: checker
    """

    def checker(hora_hand: HoraHandPattern) -> bool:
        if not isinstance(hora_hand, RegularHoraHandPattern):
            return False

        kan = 0
        for fr in hora_hand.furo:
            if isinstance(fr, Kan):
                kan += 1

        return kan == kan_count

    return checker


def sangen_series_checker_factory(sangen_kotsu_count: int, sangen_jyantou: bool):
    """
    三元系列checker工厂（小三元、大三元）
    :param sangen_kotsu_count: 三元牌刻子数
    :param sangen_jyantou: True 必须含有三元牌雀头，False 必须不含有三元牌雀头
    :return: checker
    """

    def checker(hora_hand: HoraHandPattern) -> bool:
        if not isinstance(hora_hand, RegularHoraHandPattern):
            return False

        sangen_kotsu = 0
        for kt in hora_hand.kotsu:
            if is_sangen(kt.tile):
                sangen_kotsu += 1
        return sangen_kotsu == sangen_kotsu_count and (
                sangen_jyantou and is_sangen(hora_hand.jyantou) or
                not sangen_jyantou and not is_sangen(hora_hand.jyantou)
        )

    return checker


def sushi_series_checker_factory(wind_kotsu_count: int, wind_jyantou: bool):
    """
    四喜系列checker工厂（小四喜、大四喜）
    :param wind_kotsu_count: 风牌刻子数
    :param wind_jyantou: True 必须含有风牌雀头，False 必须不含有风牌雀头
    :return: checker
    """

    def checker(hora_hand: HoraHandPattern) -> bool:
        if not isinstance(hora_hand, RegularHoraHandPattern):
            return False

        wind_kotsu = 0
        for kt in hora_hand.kotsu:
            if is_wind(kt.tile):
                wind_kotsu += 1
        return wind_kotsu == wind_kotsu_count and (
                wind_jyantou and is_wind(hora_hand.jyantou) or
                not wind_jyantou and not is_wind(hora_hand.jyantou)
        )

    return checker


def churen_series_checker_factory(nine_waiting: bool):
    """
    九莲系列checker工厂（九莲、纯九）
    :param nine_waiting: 是否纯九
    :return: checker
    """

    def checker(hora_hand: HoraHandPattern) -> bool:
        if not isinstance(hora_hand, RegularHoraHandPattern):
            return False

        type_found = [False] * 4
        num_cnt = [0] * 9

        for t in hora_hand.tiles:
            type_found[tile_type_index_mapping[t.tile_type]] = True
            num_cnt[t.real_num - 1] += 1

        type_cnt = 0
        for i in range(3):
            if type_found[i]:
                type_cnt += 1

        if type_cnt != 1 or type_found[3]:
            return False

        num_cnt[0] -= 2
        num_cnt[8] -= 2

        for i in range(9):
            num_cnt[i] -= 1
            if num_cnt[i] < 0:
                return False

        even = -1
        for i in range(9):
            if num_cnt[i] == 1:
                even = i

        return nine_waiting and hora_hand.agari.num == even + 1 or not nine_waiting and hora_hand.agari.num != even + 1

    return checker



from typing import Set

from mahjong_utils.models.hora_hand_pattern import HoraHandPattern, RegularHoraHandPattern, ChitoiHoraHandPattern
from mahjong_utils.models.tatsu import Toitsu
from mahjong_utils.models.tile import tile, is_yaochu
from mahjong_utils.models.tile_type import TileType
from mahjong_utils.yaku import _yaku, Yaku
from mahjong_utils.yaku.checker_factory import kantsu_series_checker_factory, yakuhai_checker_factory, \
    peko_series_checker_factory, anko_series_checker_factory, yaochu_series_checker_factory, \
    sangen_series_checker_factory, itsu_series_checker_factory


@_yaku(1, 1)
def tsumo(hora_hand: HoraHandPattern) -> bool:
    return hora_hand.menzen and hora_hand.tsumo


@_yaku(1, 1)
def pinhu(hora_hand: HoraHandPattern) -> bool:
    if not isinstance(hora_hand, RegularHoraHandPattern):
        return False
    if hora_hand.tsumo:
        return hora_hand.hu == 20
    else:
        return hora_hand.hu == 30


@_yaku(1, 0)
def tanyao(hora_hand: HoraHandPattern) -> bool:
    for t in hora_hand.tiles:
        if is_yaochu(t):
            return False
    return True


ipe = Yaku("ipe", 1, 1, checker=peko_series_checker_factory(1))


@_yaku(1, 0)
def self_wind(hora_hand: HoraHandPattern) -> bool:
    if not isinstance(hora_hand, RegularHoraHandPattern) or hora_hand.self_wind is None:
        return False

    for mt in hora_hand.mentsu:
        if mt.tile == hora_hand.self_wind.tile:
            return True
    return False


@_yaku(1, 0)
def round_wind(hora_hand: HoraHandPattern) -> bool:
    if not isinstance(hora_hand, RegularHoraHandPattern) or hora_hand.round_wind is None:
        return False

    for mt in hora_hand.mentsu:
        if mt.tile == hora_hand.round_wind.tile:
            return True
    return False


haku = Yaku("haku", 1, 0, checker=yakuhai_checker_factory(tile(TileType.Z, 5)))

hatsu = Yaku("hatsu", 1, 0, checker=yakuhai_checker_factory(tile(TileType.Z, 6)))

chun = Yaku("chun", 1, 0, checker=yakuhai_checker_factory(tile(TileType.Z, 7)))


@_yaku(2, 1)
def sanshoku(hora_hand: HoraHandPattern) -> bool:
    if not isinstance(hora_hand, RegularHoraHandPattern):
        return False

    shuntsu = list(hora_hand.shuntsu)

    for i in range(len(shuntsu)):
        for j in range(i + 1, len(shuntsu)):
            for k in range(j + 1, len(shuntsu)):
                x = shuntsu[i]
                y = shuntsu[j]
                z = shuntsu[k]

                if x.tile.tile_type != y.tile.tile_type \
                        and y.tile.tile_type != z.tile.tile_type \
                        and z.tile.tile_type != x.tile.tile_type \
                        and x.tile.real_num == y.tile.real_num == z.tile.real_num:
                    return True

    return False


@_yaku(2, 1)
def ittsu(hora_hand: HoraHandPattern) -> bool:
    if not isinstance(hora_hand, RegularHoraHandPattern):
        return False

    shuntsu = list(hora_hand.mentsu)

    for i in range(len(shuntsu)):
        for j in range(i + 1, len(shuntsu)):
            for k in range(j + 1, len(shuntsu)):
                x = shuntsu[i]
                y = shuntsu[j]
                z = shuntsu[k]

                if x.tile.tile_type == y.tile.tile_type == z.tile.tile_type:
                    nums = {x.tile.real_num, y.tile.real_num, z.tile.real_num}
                    if nums == {1, 4, 7}:
                        return True

    return False


chanta = Yaku("chanta", 2, 1, checker=yaochu_series_checker_factory(True, True))


@_yaku(2, 2)
def chitoi(hora_hand: HoraHandPattern) -> bool:
    return isinstance(hora_hand, ChitoiHoraHandPattern)


@_yaku(2, 0)
def toitoi(hora_hand: HoraHandPattern) -> bool:
    if not isinstance(hora_hand, RegularHoraHandPattern):
        return False

    if len(list(hora_hand.shuntsu)) != 0:
        return False

    return not hora_hand.menzen or (isinstance(hora_hand.agari_tatsu, Toitsu) and not hora_hand.tsumo)  # 非四暗刻的情况


sananko = Yaku("sananko", 2, 0, checker=anko_series_checker_factory(3))

honroto = Yaku("honroto", 2, 0, checker=yaochu_series_checker_factory(False, True))


@_yaku(2, 0)
def sandoko(hora_hand: HoraHandPattern) -> bool:
    if not isinstance(hora_hand, RegularHoraHandPattern):
        return False

    kotsu = list(hora_hand.kotsu)

    for i in range(len(kotsu)):
        for j in range(i + 1, len(kotsu)):
            for k in range(j + 1, len(kotsu)):
                x = kotsu[i]
                y = kotsu[j]
                z = kotsu[k]

                if x.tile.tile_type != y.tile.tile_type \
                        and y.tile.tile_type != z.tile.tile_type \
                        and z.tile.tile_type != x.tile.tile_type \
                        and x.tile.real_num == y.tile.real_num == z.tile.real_num:
                    return True

    return False


sankantsu = Yaku("sankantsu", 2, 0, checker=kantsu_series_checker_factory(3))

shosangen = Yaku("shosangen", 2, 0, checker=sangen_series_checker_factory(2, True))

honitsu = Yaku("honitsu", 3, 1, checker=itsu_series_checker_factory(True))

junchan = Yaku("junchan", 3, 1, checker=yaochu_series_checker_factory(True, False))

ryanpe = Yaku("ryanpe", 3, 3, checker=peko_series_checker_factory(2))

chinitsu = Yaku("chinitsu", 6, 1, checker=itsu_series_checker_factory(False))

all_common_yaku: Set[Yaku] = {
    tsumo, pinhu, tanyao, ipe, self_wind, round_wind, haku, hatsu, chun,
    sanshoku, ittsu, chanta, chitoi, toitoi, sananko, honroto, sandoko, sankantsu, shosangen,
    honitsu, junchan, ryanpe, chinitsu
}

__all__ = (
    "all_common_yaku",
    "tsumo", "pinhu", "tanyao", "ipe", "self_wind", "round_wind", "haku", "hatsu", "chun",
    "sanshoku", "ittsu", "chanta", "chitoi", "toitoi", "sananko", "honroto", "sandoko", "sankantsu", "shosangen",
    "honitsu", "junchan", "ryanpe", "chinitsu"
)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['mahjong_utils',
 'mahjong_utils.internal',
 'mahjong_utils.internal.utils',
 'mahjong_utils.models',
 'mahjong_utils.yaku']

package_data = \
{'': ['*']}

install_requires = \
['lazy>=1.5,<2.0', 'pydantic>=1.9.0,<2.0.0']

setup_kwargs = {
    'name': 'mahjong-utils',
    'version': '0.1.1a4',
    'description': '',
    'long_description': 'mahjong-utils\n========\n\n已实现功能：\n\n- [x] 获取番符对应和牌点数\n- [x] 向听数、进张分析\n- [x] 和了分析（役种、番数、符数）\n\n## 获取番符对应和牌点数\n\n```python\nfrom mahjong_utils.point_by_han_hu import get_parent_point_by_han_hu, get_child_point_by_han_hu\n\n# 获取亲家X番Y符的点数，返回(荣和点数, 自摸各家点数)\nparent_ron, parent_tsumo = get_parent_point_by_han_hu(3, 40)\n# parent_ron == 7700\n# parent_tsumo == 2600\n\n# 获取子家X番Y符的点数，返回(荣和点数, 自摸庄家点数, 自摸闲家点数)\nchild_ron, child_tsumo_parent, child_tsumo_child = get_child_point_by_han_hu(3, 40)\n# child_ron == 5200\n# child_tsumo_child == 1300\n# child_tsumo_parent == 2600\n```\n\n## 向听数、进张分析\n\n```python\nfrom mahjong_utils.models.tile import parse_tiles\nfrom mahjong_utils.shanten import shanten\n\n# 向听数、进张分析（未摸牌状态）\nresult = shanten(parse_tiles("34568m235p68s"))\n# result.shanten == 2\n# result.advance == {3M, 6M, 7M, 8M, 1P, 2P, 3P, 4P, 5P, 6S, 7S, 8S}\n\n# 向听数、进张分析（已摸牌状态）\nresult = shanten(parse_tiles("34568m235p368s"))\n# result.shanten == 2\n# result.discard_to_advance == {\n#   5P: {3M, 6M, 7M, 8M, 1P, 2P, 3P, 4P, 3S, 6S, 7S, 8S},\n#   3S: {3M, 6M, 7M, 8M, 1P, 2P, 3P, 4P, 5P, 6S, 7S, 8S},\n#   2P: {3M, 6M, 7M, 8M, 3P, 4P, 5P, 3S, 6S, 7S, 8S},\n#   8M: {3M, 6M, 1P, 2P, 4P, 5P, 3S, 7S},\n#   3M: {8M, 1P, 2P, 4P, 5P, 3S, 7S},\n#   6M: {8M ,1P, 2P, 4P, 5P, 3S, 7S},\n#   6S: {7M, 1P, 2P, 4P, 5P, 3S, 8S},\n#   8S: {7M, 1P, 2P, 4P, 5P, 3S, 6S},\n#   3P: {7M, 2P, 5P, 3S, 7S},\n# }\n```\n\n## 和了分析\n\n```python\nfrom mahjong_utils.hora import build_hora\nfrom mahjong_utils.models.tile import parse_tiles, tile\nfrom mahjong_utils.models.wind import Wind\nfrom mahjong_utils.models.furo import parse_furo\nfrom mahjong_utils.yaku.common import self_wind, round_wind\n\n# 和了分析\nhora = build_hora(\n    tiles=parse_tiles("12233466m11z"),\n    furo=[parse_furo("789p")],\n    agari=tile("1z"),\n    tsumo=True,\n    dora=4,\n    self_wind=Wind.east,\n    round_wind=Wind.east\n)\n\n# hora.yaku == {self_wind, round_wind}\n# hora.han == 6\n# hora._tiles.hu == 30\n# hora.parent_point == (18000, 6000)\n# hora.child_point == (12000, 6000, 3000)\n```',
    'author': 'ssttkkl',
    'author_email': 'huang.wen.long@hotmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from typing import Set, Dict, Optional

from pydantic import BaseModel, Field

from mahjong_utils.models.tile import Tile


class Shanten(BaseModel):
    shanten: int


class ShantenWithoutGot(Shanten):
    advance: Set[Tile] = Field(default_factory=set)
    advance_num: Optional[int]
    well_shape_advance: Optional[Set[Tile]]
    well_shape_advance_num: Optional[int]


class ShantenWithGot(Shanten):
    discard_to_advance: Dict[Tile, ShantenWithoutGot] = Field(default_factory=dict)


class ShantenInfoMixin:
    shanten_info: Shanten

    @property
    def shanten(self) -> Optional[int]:
        return getattr(self.shanten_info, "shanten", None)

    @property
    def advance(self) -> Optional[Set[Tile]]:
        return getattr(self.shanten_info, "advance", None)

    @property
    def advance_num(self) -> Optional[int]:
        return getattr(self.shanten_info, "advance_num", None)

    @property
    def well_shape_advance(self) -> Optional[Set[Tile]]:
        return getattr(self.shanten_info, "well_shape_advance", None)

    @property
    def well_shape_advance_num(self) -> Optional[int]:
        return getattr(self.shanten_info, "well_shape_advance_num", None)

    @property
    def discard_to_advance(self) -> Optional[Dict[Tile, ShantenWithoutGot]]:
        return getattr(self.shanten_info, "discard_to_advance", None)

    @property
    def with_got(self) -> bool:
        return self.discard_to_advance is not None



from collections import defaultdict
from itertools import chain
from typing import List, Dict, Literal, Optional, Sequence, Tuple, Iterable, Callable, TypeVar

from pydantic import BaseModel

from mahjong_utils.internal.hand_utils import calc_regular_advance, regular_pattern_after_discard, \
    regular_pattern_after_advance
from mahjong_utils.internal.legal_tiles_checker import ensure_legal_tiles
from mahjong_utils.internal.regular_hand_pattern_searcher import regular_hand_pattern_search
from mahjong_utils.models.furo import Furo
from mahjong_utils.models.hand import Hand
from mahjong_utils.models.hand_pattern import ChitoiHandPattern, KokushiHandPattern, RegularHandPattern, HandPattern
from mahjong_utils.models.shanten import ShantenWithGot, ShantenWithoutGot, Shanten, ShantenInfoMixin
from mahjong_utils.models.tile import Tile, is_yaochu, all_yaochu


class ShantenResult(BaseModel, ShantenInfoMixin):
    type: Literal['regular', 'chitoi', 'kokushi', 'union']
    hand: Hand
    shanten_info: Shanten


class UnionShantenResult(ShantenResult):
    regular: ShantenResult
    chitoi: Optional[ShantenResult]
    kokushi: Optional[ShantenResult]


def calc_regular_shanten(pattern: RegularHandPattern):
    shanten = 2 * (pattern.k - len(pattern.menzen_mentsu) - len(pattern.furo)) - len(pattern.tatsu)
    if pattern.jyantou is not None:
        shanten -= 1
    return shanten


def calc_chitoi_shanten(pattern: ChitoiHandPattern):
    return 6 - len(pattern.pairs)


def calc_kokushi_shanten(pattern: KokushiHandPattern):
    if pattern.repeated is not None:
        # 非十三面
        shanten_num = 12 - len(pattern.yaochu)
    else:
        # 十三面
        shanten_num = 13 - len(pattern.yaochu)
    return shanten_num


def _fill_advance_num_by_remaining(shanten_info: ShantenWithoutGot, remaining: dict):
    advance_num = 0
    for t in shanten_info.advance:
        advance_num += remaining[t]
    shanten_info.advance_num = advance_num

    if shanten_info.well_shape_advance is not None:
        well_shape_advance_num = 0
        for t in shanten_info.well_shape_advance:
            well_shape_advance_num += remaining[t]
        shanten_info.well_shape_advance_num = well_shape_advance_num


def _fill_advance_num(
        shanten_info: Shanten,
        *tiles: Tile,
):
    remaining = defaultdict(lambda: 4)
    for t in tiles:
        remaining[t] -= 1

    if isinstance(shanten_info, ShantenWithoutGot):
        _fill_advance_num_by_remaining(shanten_info, remaining)
    elif isinstance(shanten_info, ShantenWithGot):
        for shanten_info_after_discard in shanten_info.discard_to_advance.values():
            _fill_advance_num_by_remaining(shanten_info_after_discard, remaining)


T_HP = TypeVar("T_HP", bound=HandPattern)


def _select_best_patterns(
        patterns: Iterable[T_HP],
        calc_shanten: Callable[[T_HP], int]
) -> Tuple[int, Sequence[T_HP]]:
    best_shanten = 100
    best_patterns = []

    for pat in patterns:
        pat_shanten = calc_shanten(pat)
        if pat_shanten < best_shanten:
            best_shanten = pat_shanten
            best_patterns = []
        if pat_shanten == best_shanten:
            best_patterns.append(pat)

    return best_shanten, best_patterns


# ======== 标准形 ========
def _handle_regular_shanten_without_got(
        patterns: Iterable[RegularHandPattern],
        calc_well_shape_advance: bool = True
) -> Tuple[ShantenWithoutGot, Sequence[RegularHandPattern]]:
    best_shanten, best_patterns = _select_best_patterns(patterns, calc_regular_shanten)

    advance = set()
    for pat in best_patterns:
        advance |= calc_regular_advance(pat)

    shanten_info = ShantenWithoutGot(shanten=best_shanten, advance=advance)
    if calc_well_shape_advance and shanten_info.shanten == 1:
        # 一向听的牌计算好型进张
        well_shape = set()

        for adv in shanten_info.advance:
            patterns_after_adv = set()
            for pat in best_patterns:
                for pat_after_discard in regular_pattern_after_advance(pat, adv):
                    patterns_after_adv.add(pat_after_discard)

            shanten_after_adv, _ = _handle_regular_shanten_with_got(patterns_after_adv,
                                                                    calc_well_shape_advance=False,
                                                                    best_shanten_only=True)
            _fill_advance_num(shanten_after_adv, adv, *best_patterns[0].tiles,
                              *chain(*map(lambda fr: fr.tiles, best_patterns[0].furo)))

            max_adv_after_adv = max(map(lambda x: x.advance_num, shanten_after_adv.discard_to_advance.values()))
            if max_adv_after_adv > 4:
                well_shape.add(adv)

        shanten_info.well_shape_advance = well_shape

    return shanten_info, best_patterns


def _handle_regular_shanten_with_got(
        patterns: Iterable[RegularHandPattern],
        calc_well_shape_advance: bool = True,
        best_shanten_only: bool = False,
) -> Tuple[ShantenWithGot, Sequence[RegularHandPattern]]:
    best_shanten, best_patterns = _select_best_patterns(patterns, calc_regular_shanten)

    discard_to_advance = dict()
    for discard in set(best_patterns[0].tiles):
        patterns_after_discard = set()
        for pat in best_patterns:
            for pat_after_discard in regular_pattern_after_discard(pat, discard):
                patterns_after_discard.add(pat_after_discard)

        shanten_after_discard, _ = _handle_regular_shanten_without_got(patterns_after_discard, calc_well_shape_advance)
        discard_to_advance[discard] = shanten_after_discard

    if best_shanten_only:
        for t in list(discard_to_advance.keys()):
            if discard_to_advance[t].shanten != best_shanten:
                del discard_to_advance[t]

    shanten_info = ShantenWithGot(shanten=best_shanten, discard_to_advance=discard_to_advance)
    return shanten_info, best_patterns


def regular_shanten(
        tiles: Sequence[Tile],
        furo: Optional[Sequence[Furo]] = None,
        calc_advance_num: bool = True
) -> ShantenResult:
    tiles = ensure_legal_tiles(tiles)

    if furo is None:
        furo = []

    patterns = regular_hand_pattern_search(tiles, furo)

    with_got = len(tiles) % 3 == 2
    if with_got:
        shanten_info, best_patterns = _handle_regular_shanten_with_got(patterns)
    else:
        shanten_info, best_patterns = _handle_regular_shanten_without_got(patterns)

    if calc_advance_num:
        _fill_advance_num(shanten_info, *best_patterns[0].tiles,
                          *chain(*map(lambda fr: fr.tiles, best_patterns[0].furo)))

    hand = Hand(tiles=tiles, furo=furo, patterns=best_patterns)
    result = ShantenResult(type="regular", hand=hand, shanten_info=shanten_info)
    return result


# ======== 七对子 ========
def _build_chitoi_pattern(tiles: Iterable[Tile]):
    cnt: Dict[Tile, int] = {}
    for t in tiles:
        cnt[t] = cnt.get(t, 0) + 1

    pairs = []
    remaining = []
    for t in cnt:
        if cnt[t] >= 2:
            pairs.append(t)
            for i in range(cnt[t] - 2):
                remaining.append(t)
        elif cnt[t] == 1:
            remaining.append(t)

    pattern = ChitoiHandPattern(
        pairs=frozenset(pairs),
        remaining=tuple(remaining)
    )
    return pattern


def _handle_chitoi_shanten_without_got(
        tiles: Iterable[Tile]
) -> Tuple[ShantenWithoutGot, ChitoiHandPattern]:
    pattern = _build_chitoi_pattern(tiles)

    advance = set()
    for t in pattern.remaining:
        if t not in pattern.pairs:
            advance.add(t)

    shanten_num = calc_chitoi_shanten(pattern)
    shanten_info = ShantenWithoutGot(
        shanten=shanten_num,
        advance=advance
    )

    if shanten_num == 1:
        # 一向听的牌计算好型进张
        shanten_info.well_shape_advance = set()

    return shanten_info, pattern


def _handle_chitoi_shanten_with_got(
        tiles: Sequence[Tile]
) -> Tuple[ShantenWithGot, ChitoiHandPattern]:
    pattern = _build_chitoi_pattern(tiles)

    shanten_num = calc_chitoi_shanten(pattern)
    discard_to_advance = dict()

    if shanten_num != -1:
        tiles = tuple(tiles)
        for t in set(tiles):
            t_idx = tiles.index(t)

            shanten_after_discard, _ = _handle_chitoi_shanten_without_got(tiles[:t_idx] + tiles[t_idx + 1:])
            discard_to_advance[t] = shanten_after_discard

    shanten_info = ShantenWithGot(
        shanten=shanten_num,
        discard_to_advance=discard_to_advance
    )

    return shanten_info, pattern


def chitoi_shanten(
        tiles: Sequence[Tile],
        calc_advance_num: bool = True
) -> ShantenResult:
    tiles = ensure_legal_tiles(tiles, False)

    if len(tiles) == 13:
        shanten_info, pattern = _handle_chitoi_shanten_without_got(tiles)
    else:
        shanten_info, pattern = _handle_chitoi_shanten_with_got(tiles)

    if calc_advance_num:
        _fill_advance_num(shanten_info, *tiles)

    hand = Hand(tiles=tiles, patterns=[pattern])
    result = ShantenResult(type="chitoi", hand=hand, shanten_info=shanten_info)
    return result


# ======== 国士无双 ========
def _build_kokushi_pattern(tiles: Iterable[Tile]) -> Sequence[KokushiHandPattern]:
    yaochu = set()
    repeated = set()
    remaining = []

    for t in tiles:
        if is_yaochu(t):
            if t in yaochu:
                if t in repeated:
                    remaining.append(t)
                else:
                    repeated.add(t)
            else:
                yaochu.add(t)
        else:
            remaining.append(t)

    yaochu = frozenset(yaochu)
    remaining = tuple(remaining)

    patterns = []

    if len(repeated) > 0:
        # 非十三面
        for t in repeated:
            pat = KokushiHandPattern(yaochu=yaochu,
                                     repeated=t,
                                     remaining=(*remaining, *repeated.difference([t])))
            patterns.append(pat)
    else:
        # 十三面
        pat = KokushiHandPattern(yaochu=yaochu,
                                 repeated=None,
                                 remaining=remaining)
        patterns.append(pat)

    return patterns


def _handle_kokushi_shanten_without_got(
        tiles: Sequence[Tile]
) -> Tuple[ShantenWithoutGot, Sequence[KokushiHandPattern]]:
    patterns = _build_kokushi_pattern(tiles)
    shanten_num, patterns = _select_best_patterns(patterns, calc_kokushi_shanten)

    pat = patterns[0]

    if pat.repeated is not None:
        # 非十三面
        advance = all_yaochu - pat.yaochu

        if shanten_num == 1:
            well_shape_advance = set()
        else:
            well_shape_advance = None
    else:
        # 十三面
        advance = all_yaochu.copy()

        if shanten_num == 1:
            well_shape_advance = all_yaochu - pat.yaochu
        else:
            well_shape_advance = None

    shanten_info = ShantenWithoutGot(
        shanten=shanten_num,
        advance=advance,
        well_shape_advance=well_shape_advance
    )

    return shanten_info, patterns


def _handle_kokushi_shanten_with_got(
        tiles: Sequence[Tile]
) -> Tuple[ShantenWithGot, Sequence[KokushiHandPattern]]:
    patterns = _build_kokushi_pattern(tiles)

    shanten_num, patterns = _select_best_patterns(patterns, calc_kokushi_shanten)
    discard_to_advance = dict()

    if shanten_num != -1:
        tiles = tuple(tiles)

        for t in set(tiles):
            t_idx = tiles.index(t)

            shanten_after_discard, _ = _handle_kokushi_shanten_without_got(tiles[:t_idx] + tiles[t_idx + 1:])
            shanten_num = min(shanten_num, shanten_after_discard.shanten)
            discard_to_advance[t] = shanten_after_discard

    shanten_info = ShantenWithGot(
        shanten=shanten_num,
        discard_to_advance=discard_to_advance
    )

    return shanten_info, _build_kokushi_pattern(tiles)


def kokushi_shanten(
        tiles: Sequence[Tile],
        calc_advance_num: bool = True
) -> ShantenResult:
    tiles = ensure_legal_tiles(tiles, False)

    if len(tiles) == 13:
        shanten_info, patterns = _handle_kokushi_shanten_without_got(tiles)
    else:
        shanten_info, patterns = _handle_kokushi_shanten_with_got(tiles)

    if calc_advance_num:
        _fill_advance_num(shanten_info, *tiles)

    hand = Hand(tiles=tiles, patterns=patterns)
    result = ShantenResult(type="kokushi", hand=hand, shanten_info=shanten_info)
    return result


# ======== union ========
def shanten(
        tiles: List[Tile],
        furo: Optional[List[Furo]] = None,
        calc_advance_num: bool = True
) -> UnionShantenResult:
    tiles = ensure_legal_tiles(tiles)

    if furo is None:
        furo = []

    with_got = len(tiles) % 3 == 2
    k = len(tiles) // 3

    if k != 4:
        regular = regular_shanten(tiles, furo, calc_advance_num)
        return UnionShantenResult(type="union", hand=regular.hand, shanten_info=regular.shanten_info,
                                  regular=regular, chitoi=None, kokushi=None)

    regular = regular_shanten(tiles, furo, calc_advance_num=False)
    chitoi = chitoi_shanten(tiles, calc_advance_num=False)
    kokushi = kokushi_shanten(tiles, calc_advance_num=False)

    shanten_num = min(regular.shanten, chitoi.shanten, kokushi.shanten)
    patterns = []

    if not with_got:
        advance_aggregated = set()
        well_shape_advance_aggregated = set() if shanten_num == 1 else None

        if regular.shanten == shanten_num:
            advance_aggregated |= regular.advance
            patterns += regular.hand.patterns
            if shanten_num == 1:
                well_shape_advance_aggregated |= regular.well_shape_advance
        if chitoi.shanten == shanten_num:
            advance_aggregated |= chitoi.advance
            patterns += chitoi.hand.patterns
            if shanten_num == 1:
                well_shape_advance_aggregated |= chitoi.well_shape_advance
        if kokushi.shanten == shanten_num:
            advance_aggregated |= kokushi.advance
            patterns += kokushi.hand.patterns
            if shanten_num == 1:
                well_shape_advance_aggregated |= kokushi.well_shape_advance

        shanten_info = ShantenWithoutGot(
            shanten=shanten_num,
            advance=advance_aggregated,
            well_shape_advance=well_shape_advance_aggregated,
        )
    else:
        discard_to_advance_aggregated = dict()

        for discard, advance in regular.discard_to_advance.items():
            if (discard not in discard_to_advance_aggregated
                    or discard_to_advance_aggregated[discard].shanten > advance.shanten):
                discard_to_advance_aggregated[discard] = advance
            elif discard_to_advance_aggregated[discard].shanten == advance.shanten:
                discard_to_advance_aggregated[discard].advance |= advance.advance
                if discard_to_advance_aggregated[discard].shanten == 1:
                    discard_to_advance_aggregated[discard].well_shape_advance |= advance.well_shape_advance
        patterns += regular.hand.patterns

        for discard, advance in chitoi.discard_to_advance.items():
            if (discard not in discard_to_advance_aggregated
                    or discard_to_advance_aggregated[discard].shanten > advance.shanten):
                discard_to_advance_aggregated[discard] = advance
            elif discard_to_advance_aggregated[discard].shanten == advance.shanten:
                discard_to_advance_aggregated[discard].advance |= advance.advance
                if discard_to_advance_aggregated[discard].shanten == 1:
                    discard_to_advance_aggregated[discard].well_shape_advance |= advance.well_shape_advance
        patterns += chitoi.hand.patterns

        for discard, advance in kokushi.discard_to_advance.items():
            if (discard not in discard_to_advance_aggregated
                    or discard_to_advance_aggregated[discard].shanten > advance.shanten):
                discard_to_advance_aggregated[discard] = advance
            elif discard_to_advance_aggregated[discard].shanten == advance.shanten:
                discard_to_advance_aggregated[discard].advance |= advance.advance
                if discard_to_advance_aggregated[discard].shanten == 1:
                    discard_to_advance_aggregated[discard].well_shape_advance |= advance.well_shape_advance
        patterns += kokushi.hand.patterns

        shanten_info = ShantenWithGot(
            shanten=shanten_num,
            discard_to_advance=discard_to_advance_aggregated
        )

    if calc_advance_num:
        _fill_advance_num(shanten_info, *tiles, *chain(*map(lambda fr: fr.tiles, furo)))

    hand = Hand(tiles=tiles, furo=furo, patterns=patterns)
    return UnionShantenResult(type="union", hand=hand, shanten_info=shanten_info,
                              regular=regular, chitoi=chitoi, kokushi=kokushi)


__all__ = ("regular_shanten",
           "chitoi_shanten",
           "kokushi_shanten",
           "shanten",
           "ShantenResult",
           "UnionShantenResult",)



from typing import Sequence, Tuple

from mahjong_utils.models.tile import Tile, tile


def ensure_legal_tiles(tiles: Sequence[Tile], allow_any_k: bool = True) -> Tuple[Tile, ...]:
    if not allow_any_k and len(tiles) // 3 != 4:
        raise ValueError(f"invalid length of hand: {len(tiles)}")
    if len(tiles) < 1 or len(tiles) > 14 or len(tiles) % 3 == 0:
        raise ValueError(f"invalid length of hand: {len(tiles)}")

    cnt = {}
    for t in tiles:
        cnt[t] = cnt.get(t, 0) + 1
        if cnt[t] > 4:
            raise ValueError(f"invalid num of tile {t}")

    ret_tiles = tuple(tile(t.tile_type, 5) if t.num == 0 else t for t in tiles)
    return ret_tiles
